/*
** USBSlaveHIDKbd.c
**
** Copyright  2009-2010-2011 Future Technology Devices International Limited
**
** C file containing the USB Slave HID keyboard driver code.
**
** Author: FTDI
** Project: Vinculum II Drivers
** Module: Vinculum II USB Slave HID Keyboard Driver
** Requires: vos.h ioctl.h usbslave.h
** Comments:
**
** History:
**  1  Initial version
**
*/
#include "vos.h"
#include "devman.h"
#include "memmgmt.h"

#include "ioctl.h"

#include "USB.h"
#include "USBHID.h"
#include "USBSlave.h"
#include "USBSlaveHIDKbdDrv.h"

void usbslavehidkbd_setup(usbSlaveHIDKbd_context *ctx);
void usbslavehidkbd_in(usbSlaveHIDKbd_context *ctx);
unsigned char standard_request(usbSlaveHIDKbd_context *ctx);
unsigned char class_request(usbSlaveHIDKbd_context *ctx);

unsigned char usbSlaveHIDKbd_write(
	char *xfer,
	unsigned short num_to_write,
	unsigned short *num_written,
	usbSlaveHIDKbd_context *ctx);

unsigned char usbSlaveHIDKbd_ioctl(common_ioctl_cb_t *cb, usbSlaveHIDKbd_context *ctx);

// thread states
#define UNATTACHED 0
#define ATTACHED   1

/*
   Device Descriptor:
   bcdUSB:             0x0110
   bDeviceClass:         0x00
   bDeviceSubClass:      0x00
   bDeviceProtocol:      0x00
   bMaxPacketSize0:      0x08 (8)
   idVendor:           0x0403 (Future Technology Devices International Limited)
   idProduct:          0xFEED
   bcdDevice:          0x0101
   iManufacturer:        0x01
   0x0409: "FTDI"
   iProduct:             0x02
   0x0409: "USB Keyboard"
   iSerialNumber:        0x00
   bNumConfigurations:   0x01
 */
unsigned char device_descriptor[18] = {
	0x12, 0x01, 0x10, 0x01, 0x00, 0x00, 0x00, 0x08,
	0x03, 0x04, 0xed, 0xfe, 0x01, 0x01, 0x01, 0x02,0x00,  0x01
};

/*
   Configuration Descriptor:
   wTotalLength:       0x0022
   bNumInterfaces:       0x01
   bConfigurationValue:  0x01
   iConfiguration:       0x00
   bmAttributes:         0xA0 (Bus Powered Remote Wakeup)
   MaxPower:             0x32 (100 Ma)

   Interface Descriptor:
   bInterfaceNumber:     0x00
   bAlternateSetting:    0x00
   bNumEndpoints:        0x01
   bInterfaceClass:      0x03 (HID)
   bInterfaceSubClass:   0x01
   bInterfaceProtocol:   0x01
   iInterface:           0x00

   HID Descriptor:
   bcdHID:             0x0110
   bCountryCode:         0x00
   bNumDescriptors:      0x01
   bDescriptorType:      0x22
   wDescriptorLength:  0x0041

   Endpoint Descriptor:
   bEndpointAddress:     0x81  IN
   Transfer Type:   Interrupt
   wMaxPacketSize:     0x0008 (8)
   bInterval:            0x0A
 */
unsigned char config_descriptor[34] = {
	0x09, 0x02, 0x22, 0x00, 0x01, 0x01, 0x00, 0xA0, 0x32,
	0x09, 0x04, 0x00, 0x00, 0x01, 0x03, 0x01, 0x01, 0x00,
	0x09, 0x21, 0x10, 0x01, 0x00, 0x01, 0x22, 0x41, 0x00,
	0x07, 0x05, 0x81, 0x03, 0x08, 0x00, 0x0A
};

// English (US)
unsigned char str0_descriptor[4] = {
	0x04, 0x03, 0x09, 0x04
};

// "FTDI"
unsigned char str1_descriptor[10] = {
	0x0a, 0x03, 0x46, 0x00, 0x54, 0x00, 0x44, 0x00, 0x49, 0x00
};

// "USB Keyboard"
unsigned char str2_descriptor[26] = {
	0x1a, 0x03, 0x55, 0x00, 0x53, 0x00, 0x42, 0x00, 0x20, 0x00,
	0x4b, 0x00, 0x65, 0x00, 0x79, 0x00, 0x62, 0x00, 0x6f, 0x00,
	0x61, 0x00, 0x72, 0x00, 0x64, 0x00
};

// See Device Class Definition for Human Interface Devices (HID) Version 1.11
// from USB Implementers Forum USB.org
unsigned char report1_descriptor[65] = {
	0x05, 0x01, 0x09, 0x06, 0xA1, 0x01, 0x05, 0x07, 0x19, 0xE0, 0x29,
	0xE7, 0x15, 0x00, 0x25, 0x01, 0x75, 0x01, 0x95, 0x08, 0x81, 0x02,
	0x95, 0x01, 0x75, 0x08, 0x81, 0x01, 0x95, 0x03, 0x75, 0x01, 0x05,
	0x08, 0x19, 0x01, 0x29, 0x03, 0x91, 0x02, 0x95, 0x05, 0x75, 0x01,
	0x91, 0x01, 0x95, 0x06, 0x75, 0x08, 0x15, 0x00, 0x26, 0xFF, 0x00,
	0x05, 0x07, 0x19, 0x00, 0x2A, 0xFF, 0x00, 0x81, 0x00, 0xC0
};

unsigned char report2_descriptor[133] = {
	0x06, 0x01, 0x00, 0x09, 0x80, 0xA1, 0x01, 0x85, 0x01, 0x25, 0x01,
	0x15, 0x00, 0x75, 0x01, 0x0A, 0x81, 0x00, 0x0A, 0x82, 0x00, 0x0A,
	0x83, 0x00, 0x95, 0x03, 0x81, 0x06, 0x95, 0x05, 0x81, 0x01, 0xC0,
	0x06, 0x0C, 0x00, 0x09, 0x01, 0xA1, 0x01, 0x85, 0x02, 0x25, 0x01,
	0x15, 0x00, 0x75, 0x01, 0x0A, 0x8A, 0x01, 0x0A, 0x23, 0x02, 0x0A,
	0x2A, 0x02, 0x0A, 0x21, 0x02, 0x0A, 0x27, 0x02, 0x0A, 0x26, 0x02,
	0x0A, 0x25, 0x02, 0x0A, 0x24, 0x02, 0x95, 0x08, 0x81, 0x02, 0x0A,
	0x83, 0x01, 0x0A, 0xCD, 0x00, 0x0A, 0xB7, 0x00, 0x0A, 0xB6, 0x00,
	0x0A, 0xB5, 0x00, 0x0A, 0xE9, 0x00, 0x0A, 0xEA, 0x00, 0x0A, 0xE2,
	0x00, 0x95, 0x08, 0x81, 0x02, 0x0A, 0x94, 0x01, 0x0A, 0x92, 0x01,
	0x0A, 0x96, 0x01, 0x0A, 0x9E, 0x01, 0x0A, 0xB2, 0x00, 0x0A, 0xB4,
	0x00, 0x0A, 0x06, 0x02, 0x0A, 0xB8, 0x00, 0x95, 0x08, 0x81, 0x02,
	0xC0
};

unsigned char usbslavehidkbd_init(unsigned char vos_dev_num)
{
	usbSlaveHIDKbd_context *ctx;
	vos_driver_t *usbSlaveHIDKbd_cb;

	ctx = vos_malloc(sizeof(usbSlaveHIDKbd_context));

	if (ctx == NULL)
		return USBSLAVEHID_ERROR;

	usbSlaveHIDKbd_cb = vos_malloc(sizeof(vos_driver_t));

	if (usbSlaveHIDKbd_cb == NULL)
	{
		vos_free(ctx);
		return USBSLAVEHID_ERROR;
	}

	// Set up function pointers for our driver
	usbSlaveHIDKbd_cb->flags = 0;
	usbSlaveHIDKbd_cb->read = NULL;
	usbSlaveHIDKbd_cb->write = usbSlaveHIDKbd_write;
	usbSlaveHIDKbd_cb->ioctl = usbSlaveHIDKbd_ioctl;
	usbSlaveHIDKbd_cb->interrupt = (PF_INT) NULL;
	usbSlaveHIDKbd_cb->open = (PF_OPEN) NULL;
	usbSlaveHIDKbd_cb->close = (PF_CLOSE) NULL;

	// OK - register with device manager
	vos_dev_init(vos_dev_num, usbSlaveHIDKbd_cb, ctx);

	// defaults
	ctx->attached = 0;
	ctx->report = 0;

	ctx->tcbSetup = vos_create_thread_ex(31, SIZEOF_KBD_SETUP_MEMORY, usbslavehidkbd_setup, "HIDSetup", 2, ctx);
	ctx->tcbIn = vos_create_thread_ex(31, SIZEOF_KBD_IN_MEMORY, usbslavehidkbd_in, "HIDIn", 2, ctx);

	vos_init_mutex(&ctx->writeLock, 0);

	vos_init_semaphore(&ctx->writeDataAvail, 0);
	vos_init_semaphore(&ctx->reportSet, 0);

	if ((ctx->tcbSetup) && (ctx->tcbIn))
		return USBSLAVEHID_OK;

	return USBSLAVEHID_ERROR;
}

void usbslavehidkbd_setup(usbSlaveHIDKbd_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	usb_deviceRequest_t *devReq;
	unsigned char bmRequestType;
	unsigned char state = UNATTACHED;

	while (1)
	{
		switch (state)
		{
		case UNATTACHED:

			if (!ctx->attached)
				vos_delay_msecs(100);
			else
			{
				state = ATTACHED;
			}

			break;

		case ATTACHED:

			if (!ctx->attached)
			{
				state = UNATTACHED;
				break;
			}

			iocb.ioctl_code = VOS_IOCTL_USBSLAVE_WAIT_SETUP_RCVD;
			iocb.request.setup_or_bulk_transfer.buffer = ctx->setup_buffer;
			iocb.request.setup_or_bulk_transfer.size = 9;
			vos_dev_ioctl(ctx->handle, &iocb);

			devReq = (usb_deviceRequest_t *) ctx->setup_buffer;
			bmRequestType = devReq->bmRequestType & (USB_BMREQUESTTYPE_STANDARD | USB_BMREQUESTTYPE_CLASS);

			// we only need to handle standard and class requests for HID keyboard
			if (bmRequestType == USB_BMREQUESTTYPE_STANDARD)
			{
				standard_request(ctx);
			}
			else if (bmRequestType == USB_BMREQUESTTYPE_CLASS)
			{
				class_request(ctx);
			}

			break;

		default:
			asm {HALT};
			break;
		}
	}

	return;
}

void usbslavehidkbd_in(usbSlaveHIDKbd_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	unsigned char state = UNATTACHED;

	while (1)
	{
		switch (state)
		{
		case UNATTACHED:

			if (!ctx->attached)
				vos_delay_msecs(100);
			else
			{
				state = ATTACHED;
			}

			break;

		case ATTACHED:

			if (!ctx->attached)
			{
				state = UNATTACHED;
				break;
			}

			// wait for data to be avialble
			vos_wait_semaphore(&ctx->writeDataAvail);

			vos_lock_mutex(&ctx->writeLock);

			if (ctx->report == 1)
			{
				// write requests to the lower layer will be completed when the host
				// requests an IN - this will happen with the polling interval specified
				// in the endpoint descriptor (in this case 10ms)
				iocb.ioctl_code = VOS_IOCTL_USBSLAVE_TRANSFER;
				iocb.handle = ctx->in_ep;
				iocb.request.setup_or_bulk_transfer.buffer = &ctx->ep_in_buffer[0];
				iocb.request.setup_or_bulk_transfer.size = (int16) 8;
				vos_dev_ioctl(ctx->handle, &iocb);
				vos_memset(&ctx->ep_in_buffer, 0, 8);
			}

			vos_unlock_mutex(&ctx->writeLock);

			break;

		default:
			asm {HALT};
			break;
		}
	}

	return;
}

void usbSlaveHIDKbd_detach(usbSlaveHIDKbd_context *ctx)
{
	ctx->attached = 0;

	return;
}

unsigned char usbSlaveHIDKbd_attach(VOS_HANDLE handle, usbSlaveHIDKbd_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	unsigned char status = USBSLAVEHID_OK;

	// save usb slave handle
	ctx->handle = handle;

	if (!ctx->attached)
	{
		// get endpoint handles and set max packet sizes
		// these should match the descriptor values!
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_CONTROL_ENDPOINT_HANDLE;
		iocb.ep = USBSLAVE_CONTROL_IN;
		iocb.get = &ctx->in_ep0;
		vos_dev_ioctl(ctx->handle, &iocb);

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ENDPOINT_MAX_PACKET_SIZE;
		iocb.handle = ctx->in_ep0;
		iocb.request.ep_max_packet_size = USBSLAVE_MAX_PACKET_SIZE_8;
		vos_dev_ioctl(ctx->handle, &iocb);

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_CONTROL_ENDPOINT_HANDLE;
		iocb.ep = USBSLAVE_CONTROL_OUT;
		iocb.get = &ctx->out_ep0;
		vos_dev_ioctl(ctx->handle, &iocb);

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ENDPOINT_MAX_PACKET_SIZE;
		iocb.handle = ctx->out_ep0;
		iocb.request.ep_max_packet_size = USBSLAVE_MAX_PACKET_SIZE_8;
		vos_dev_ioctl(ctx->handle, &iocb);

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_INT_IN_ENDPOINT_HANDLE;
		iocb.ep = USBSLAVEHIDKBD_IN;
		iocb.get = &ctx->in_ep;
		vos_dev_ioctl(ctx->handle, &iocb);

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ENDPOINT_MAX_PACKET_SIZE;
		iocb.handle = ctx->in_ep;
		iocb.request.ep_max_packet_size = USBSLAVE_MAX_PACKET_SIZE_8;
		vos_dev_ioctl(ctx->handle, &iocb);

		ctx->attached = 1;

		// wait until we've processed a SET_REPORT request
		// after this, we can safely process data
		vos_wait_semaphore(&ctx->reportSet);
	}

	return status;
}

// write function
unsigned char usbSlaveHIDKbd_write(
	char *xfer,
	unsigned short num_to_write,
	unsigned short *num_written,
	usbSlaveHIDKbd_context *ctx)
{
	vos_lock_mutex(&ctx->writeLock);

	// 8 byte report
	vos_memcpy(&ctx->ep_in_buffer, xfer, 8);

	// Signal IN thread that we have data
	vos_signal_semaphore(&ctx->writeDataAvail);

	vos_unlock_mutex(&ctx->writeLock);

	return USBSLAVEHID_OK;
}

// USB Slave IOCTL function
unsigned char usbSlaveHIDKbd_ioctl(common_ioctl_cb_t *cb, usbSlaveHIDKbd_context *ctx)
{
	unsigned char status = USBSLAVEHID_INVALID_PARAMETER;

	switch (cb->ioctl_code)
	{
	case VOS_IOCTL_USBSLAVEHID_ATTACH:
		status = usbSlaveHIDKbd_attach((VOS_HANDLE) cb->set.data, ctx);
		break;

	case VOS_IOCTL_USBSLAVEHID_DETACH:
		usbSlaveHIDKbd_detach(ctx);
		status = USBSLAVEHID_OK;
		break;

	default:
		break;
	}

	return status;
}

void ack_request(usbSlaveHIDKbd_context *ctx)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
	iocb.handle = ctx->in_ep0;
	iocb.request.setup_or_bulk_transfer.buffer = (void *) 0;
	iocb.request.setup_or_bulk_transfer.size = 0;
	vos_dev_ioctl(ctx->handle, &iocb);
}

void set_address_request(usbSlaveHIDKbd_context *ctx, unsigned char addr)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ADDRESS;
	iocb.set = (void *) addr;
	vos_dev_ioctl(ctx->handle, &iocb);

	ack_request(ctx);
}

void set_configuration_request(usbSlaveHIDKbd_context *ctx, unsigned char config)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_CONFIGURATION;
	iocb.set = (void *) config;
	vos_dev_ioctl(ctx->handle, &iocb);

	ack_request(ctx);
}

void get_descriptor_request(usbSlaveHIDKbd_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	usb_deviceRequest_t *devReq;
	unsigned char hValue;
	unsigned char lValue;
	unsigned short wLength;
	unsigned short siz;
	uint32 ul_siz;
	unsigned char *src;
	unsigned char cond;

	devReq = (usb_deviceRequest_t *) ctx->setup_buffer;

	hValue = devReq->wValue >> 8;
	lValue = devReq->wValue & 0xff;

	wLength = devReq->wLength;

	switch (hValue)
	{
	case USB_DESCRIPTOR_TYPE_DEVICE:
		ul_siz = (uint32) wLength;
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->in_ep0;
		iocb.request.setup_or_bulk_transfer.buffer = device_descriptor;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle, &iocb);
		return;
		break;

	case USB_DESCRIPTOR_TYPE_CONFIGURATION:
		siz = wLength == 9?9:sizeof(config_descriptor);
		ul_siz = (uint32) siz;

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->in_ep0;
		iocb.request.setup_or_bulk_transfer.buffer = config_descriptor;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle, &iocb);
		return;

	case USB_DESCRIPTOR_TYPE_STRING:

		if (lValue == 0)
		{
			src = str0_descriptor;
			siz = sizeof(str0_descriptor);
		}
		else if (lValue == 1)
		{
			src = str1_descriptor;
			siz = sizeof(str1_descriptor);
		}
		else if (lValue == 2)
		{
			src = str2_descriptor;
			siz = sizeof(str2_descriptor);
		}

		cond = (unsigned char) (wLength != siz);

		if (siz > wLength)
			siz = wLength;

		ul_siz = (uint32) siz;

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->in_ep0;
		iocb.request.setup_or_bulk_transfer.buffer = src;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle, &iocb);
		return;

	case USB_DESCRIPTOR_TYPE_REPORT:

		if (lValue == 0)
		{
			src = report1_descriptor;
			siz = sizeof(report1_descriptor);
		}
		else if (lValue == 1)
		{
			src = report2_descriptor;
			siz = sizeof(report2_descriptor);
		}

		cond = (unsigned char) (wLength != siz);

		if (siz > wLength)
			siz = wLength;

		ul_siz = (uint32) siz;

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->in_ep0;
		iocb.request.setup_or_bulk_transfer.buffer = src;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle, &iocb);
		return;
		break;

	default:
		// respond with Request Error
		return;
	}
}

void set_feature_request(usbSlaveHIDKbd_context *ctx, usbslave_ep_handle_t ep)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_STALL;
	iocb.ep = ep;
	vos_dev_ioctl(ctx->handle, &iocb);

	ack_request(ctx);
}

void clear_feature_request(usbSlaveHIDKbd_context *ctx, usbslave_ep_handle_t ep)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_CLEAR;
	iocb.ep = ep;
	vos_dev_ioctl(ctx->handle, &iocb);

	ack_request(ctx);
}

void get_ep_status(usbSlaveHIDKbd_context *ctx, usbslave_ep_handle_t ep)
{
	usbslave_ioctl_cb_t iocb;
	char state;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_STATE;
	iocb.ep = ep;
	iocb.get = &state;
	vos_dev_ioctl(ctx->handle, &iocb);

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
	iocb.handle = ctx->out_ep0;
	iocb.request.setup_or_bulk_transfer.buffer = (void *) &state;
	iocb.request.setup_or_bulk_transfer.size = 1;
	vos_dev_ioctl(ctx->handle, &iocb);
}

void set_control_ep_halt(usbSlaveHIDKbd_context *ctx)
{
	usbslave_ioctl_cb_t iocb;

	// Performs a protocol stall on endpoint 0
	// Indicates that a request is unsupported
	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_STALL;
	iocb.ep = 0;
	vos_dev_ioctl(ctx->handle, &iocb);
}

unsigned char standard_request(usbSlaveHIDKbd_context *ctx)
{
	usb_deviceRequest_t *devReq;
	unsigned char status = USBSLAVE_OK;
	unsigned char bReq;

	devReq = (usb_deviceRequest_t *) ctx->setup_buffer;
	bReq = devReq->bRequest;

	switch (bReq)
	{
	case USB_REQUEST_CODE_SET_ADDRESS:
		set_address_request(ctx, devReq->wValue & 0xff);
		break;

	case USB_REQUEST_CODE_GET_DESCRIPTOR:
		get_descriptor_request(ctx);
		break;

	case USB_REQUEST_CODE_SET_CONFIGURATION:
		set_configuration_request(ctx, devReq->wValue & 0xff);
		break;

	case USB_REQUEST_CODE_CLEAR_FEATURE:
		clear_feature_request(ctx, devReq->wIndex >> 8);
		break;

	case USB_REQUEST_CODE_SET_FEATURE:
		set_feature_request(ctx, devReq->wIndex >> 8);
		break;

	default:
		// force a protocol stall
		set_control_ep_halt(ctx);
		break;
	}

	return status;
}

void class_ack(usbSlaveHIDKbd_context *ctx)
{
	ack_request(ctx);
}

void class_control_out(usbSlaveHIDKbd_context *ctx, char *buffer, unsigned short len)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
	iocb.handle = ctx->out_ep0;
	iocb.request.setup_or_bulk_transfer.buffer = (void *) buffer;
	iocb.request.setup_or_bulk_transfer.size = len;
	vos_dev_ioctl(ctx->handle, &iocb);
}

unsigned char class_request(usbSlaveHIDKbd_context *ctx)
{
	usb_deviceRequest_t *devReq;
	usbslave_ioctl_cb_t iocb;
	unsigned char status = USBSLAVE_OK;
	unsigned char bReq;

	devReq = (usb_deviceRequest_t *) ctx->setup_buffer;
	bReq = devReq->bRequest;

	switch (bReq)
	{
	case USB_HID_REQUEST_CODE_SET_IDLE:
		class_ack(ctx);
		break;

	case USB_HID_REQUEST_CODE_SET_PROTOCOL:
		class_ack(ctx);
		break;

	case USB_HID_REQUEST_CODE_SET_REPORT:
		// dummy read of one byte
		class_control_out(ctx, (signed) &status, 1);
		ctx->report = 1;
		// allow attach to return now that we can safely accept data
		vos_signal_semaphore(&ctx->reportSet);
		break;

	default:
		// force a protocol stall
		set_control_ep_halt(ctx);
		break;
	}

	return status;
}

